from LigParGen.BOSSReader import BOSSReader, CheckForHs
from LigParGen.BOSS2OPENMM import mainBOSS2OPM
from LigParGen.BOSS2CHARMM import mainBOSS2CHARMM
from LigParGen.BOSS2GMX import mainBOSS2GMX
from LigParGen.BOSS2XPLOR import mainBOSS2XPLOR
from LigParGen.BOSS2Q import mainBOSS2Q
from LigParGen.BOSS2LAMMPS import mainBOSS2LAMMPS
from LigParGen.BOSS2DESMOND import mainBOSS2DESMOND 
from LigParGen.CreatZmat import GenMolRep
import argparse
import pickle
import os
# from shutil import which

def main():

    parser = argparse.ArgumentParser(
        prog='LigParGen',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
    Ligand Parameter Generator Based on 
    Jorgensen group's OPLS-AA/CM1A(-LBCC) FF
    Created on Mon Feb 15 15:40:05 2016
    @author: Leela S. Dodda leela.dodda@yale.edu
    @author: William L. Jorgensen Lab 

    FF formats provided : 
    --------------------
    OpenMM       .xml  
    CHARMM/NAMD  .prm & .rtf  
    GROMACS      .itp & .gro 
    CNS/X-PLOR   .param & .top
    Q            .Q.prm & .Q.lib
    DESMOND      .cms
    BOSS/MCPRO   .z
    PDB2PQR      .pqr

    Input Files supported : 
    --------------------
    SMILES code
    PDB
    MDL MOL Format

    ################################################ 
    if using MOL file 
    Usage: LigParGen -m phenol.mol    -r PHN -c 0 -o 0

    if using PDB file 
    Usage: LigParGen -p phenol.pdb    -r PHN -c 0 -o 0
    
    if using BOSS SMILES CODE 
    Usage: LigParGen -s 'c1ccc(cc1)O' -r PHN -c 0 -o 0  
    
    REQUIREMENTS:
    BOSS (need to set BOSSdir in bashrc and cshrc)
    Preferably Anaconda python with following modules
    pandas 
    argparse
    numpy
    openbabel

    Please cite following references: 
    1. LigParGen web server: an automatic OPLS-AA parameter generator for organic ligands  
       Leela S. Dodda  Israel Cabeza de Vaca  Julian Tirado-Rives William L. Jorgensen 
       Nucleic Acids Research, Volume 45, Issue W1, 3 July 2017, Pages W331–W336
    2. 1.14*CM1A-LBCC: Localized Bond-Charge Corrected CM1A Charges for Condensed-Phase Simulations
       Leela S. Dodda, Jonah Z. Vilseck, Julian Tirado-Rives , and William L. Jorgensen 
       Department of Chemistry, Yale University, New Haven, Connecticut 06520-8107, United States
       J. Phys. Chem. B, 2017, 121 (15), pp 3864–3870
    3. Accuracy of free energies of hydration using CM1 and CM3 atomic charges.
       Udier–Blagović, M., Morales De Tirado, P., Pearlman, S. A. and Jorgensen, W. L. 
       J. Comput. Chem., 2004, 25,1322–1332. doi:10.1002/jcc.20059
    """
    )
    parser.add_argument(
        "-r", "--resname", help="Residue name from PDB FILE", type=str)
    parser.add_argument(
        "-s", "--smiles", help="Paste SMILES code from CHEMSPIDER or PubChem", type=str)
    parser.add_argument(
        "-m", "--mol", help="Submit MOL file from CHEMSPIDER or PubChem", type=str)
    parser.add_argument(
        "-p", "--pdb", help="Submit PDB file from CHEMSPIDER or PubChem", type=str)
    parser.add_argument(
        "-o", "--opt", help="Optimization or Single Point Calculation", type=int, choices=[0, 1, 2, 3])
    parser.add_argument("-c", "--charge", type=int,
                        choices=[0, -1, 1, -2, 2], help="0: Neutral <0: Anion >0: Cation ")
    parser.add_argument(
        "-l", "--lbcc", help="Use 1.14*CM1A-LBCC charges instead of 1.14*CM1A", action="store_true")
    args = parser.parse_args()

    convert(**vars(args))

def convert(**kwargs):

    # set the default values
    options = {
            'opt' : 0,
            'smiles' : None,
            'zmat' : None, 
            'charge' : 0,
            'lbcc' : False,
            'mol' : None,
            'resname' : 'UNK',
            'pdb' : None }

    # update the default values based on the arguments
    options.update(kwargs)

    # set the arguments that you would used to get from argparse
    opt = options['opt']
    smiles = options['smiles']
    zmat = options['zmat']
    charge = options['charge']
    lbcc = options['lbcc']
    resname = options['resname']
    mol = options['mol']
    pdb = options['pdb']

    if opt != None:
        optim = opt
    else:
        optim = 0

    clu = False
    #charge = charge
    lbcc = False

    # assert (which('obabel')
            # is not None), "OpenBabel is Not installed or \n the executable location is not accessable"
    if os.path.exists('/tmp/' + resname + '.xml'):
        os.system('/bin/rm /tmp/' + resname + '.*')
    if lbcc:
        if charge == 0:
            lbcc = True
        else:
            lbcc = False
            print(
                '1.14*CM1A-LBCC is only available for neutral molecules\n Assigning unscaled CM1A charges')

    if smiles != None:
        os.chdir('/tmp/')
        smifile = open('%s.smi' % resname, 'w+')
        smifile.write('%s' % smiles)
        smifile.close()
        GenMolRep('%s.smi' % resname, optim, resname, charge)
        mol = BOSSReader('%s.z' % resname, optim, charge, lbcc)
    elif mol != None:
        os.system('cp %s /tmp/' % mol)
        os.chdir('/tmp/')
        GenMolRep(mol, optim, resname, charge)
        mol = BOSSReader('%s.z' % resname, optim, charge, lbcc)
    elif pdb != None:
        os.system('cp %s /tmp/' % pdb)
        os.chdir('/tmp/')
        GenMolRep(pdb, optim, resname, charge)
        mol = BOSSReader('%s.z' % resname, optim, charge, lbcc)
        clu = True
    assert (mol.MolData['TotalQ']['Reference-Solute'] ==
            charge), "PROPOSED CHARGE IS NOT POSSIBLE: SOLUTE MAY BE AN OPEN SHELL"
    assert(CheckForHs(mol.MolData['ATOMS'])
           ), "Hydrogens are not added. Please add Hydrogens"

    pickle.dump(mol, open(resname + ".p", "wb"))
    mainBOSS2OPM(resname, clu)
    print('DONE WITH OPENMM')
    mainBOSS2Q(resname, clu)
    print('DONE WITH Q')
    mainBOSS2XPLOR(resname, clu)
    print('DONE WITH XPLOR')
    mainBOSS2CHARMM(resname, clu)
    print('DONE WITH CHARMM/NAMD')
    mainBOSS2GMX(resname, clu)
    print('DONE WITH GROMACS')
    mainBOSS2LAMMPS(resname, clu)
    print('DONE WITH LAMMPS')
    mainBOSS2DESMOND(resname, clu)
    print('DONE WITH DESMOND')
    os.remove(resname + ".p")
    mol.cleanup()

if __name__ == "__main__":
  
    main()



"""
SCRIPT TO WRITE DESMOND CMS FILES 
FROM BOSS ZMATRIX
Created on Mon Nov 26 15:40:05 2017
@author: Leela S. Dodda leela.dodda@yale.edu
@author: William L. Jorgensen Lab 

Usage: python OPM_Routines.py -z phenol.z -r PHN
REQUIREMENTS:
BOSS (need to set BOSSdir in bashrc and cshrc)
Preferably Anaconda python with following modules
pandas 
argparse
numpy
"""

from LigParGen.BOSSReader import ucomb,pairing_func,bossPdbAtom2Element,bossElement2Mass
from LigParGen.BOSSReader import Refine_PDB_file,get_coos_from_pdb
import pickle
import pandas as pd
import numpy as np


def boss2opmTorsion(bnd_df, num2opls, st_no, molecule_data, itpf):
    dhd = []
    for line in molecule_data.MolData['TORSIONS']:
        dt = [float(f) for f in line]
        dhd.append(dt)
    dhd = np.array(dhd)
    dhd = dhd * 1.  # kcal to kj conversion
    dhd = dhd  # Komm = Vopls/2
    dhd_df = pd.DataFrame(dhd, columns=['V1', 'V2', 'V3', 'V4'])
    ats = []
    for line in molecule_data.MolData['ATOMS'][3:]:
        dt = [line.split()[0], line.split()[4],
              line.split()[6], line.split()[8]]
        dt = [int(d) for d in dt]
        ats.append(dt)
    for line in molecule_data.MolData['ADD_DIHED']:
        dt = [int(l) for l in line]
        ats.append(dt)
    assert len(ats) == len(
        dhd), 'Number of Dihedral angles in Zmatrix and Out file dont match'
    ats = np.array(ats) - st_no
    for i in range(len(ats)):
        for j in range(len(ats[0])):
            if ats[i][j] < 0:
                ats[i][j] = 0
    at_df = pd.DataFrame(ats, columns=['I', 'J', 'K', 'L'])
    final_df = pd.concat([dhd_df, at_df], axis=1, join_axes=[at_df.index])
    bndlist = list(bnd_df.UR) + (list(bnd_df.UR))
    final_df['TY'] = ['Proper' if ucomb(list([final_df.I[n], final_df.J[n], final_df.K[
        n], final_df.L[n]]), bndlist) == 3 else 'Improper' for n in range(len(final_df.I))]
    final_df['SumV'] = np.abs(
        final_df.V1) + np.abs(final_df.V2) + np.abs(final_df.V3) + np.abs(final_df.V4)
    final_df['TI'] = [num2opls[j] for j in final_df.I]
    final_df['TJ'] = [num2opls[j] for j in final_df.J]
    final_df['TK'] = [num2opls[j] for j in final_df.K]
    final_df['TL'] = [num2opls[j] for j in final_df.L]
    if len(final_df.index) > 0:
        final_df['NAME'] = final_df.TI + '-' + final_df.TJ + \
            '-' + final_df.TK + '-' + final_df.TL
        final_df = final_df.sort_values(['NAME'])
        tor_bos = final_df.drop(
            ['I', 'J', 'K', 'L', 'TI', 'TJ', 'TK', 'TL'], 1)
        tor_bos = tor_bos.drop_duplicates()
        df = final_df.ix[tor_bos.index]
        return final_df, df
    else:
        return final_df, final_df


def boss2gmxBond(molecule_data, st_no, itpf):
    bdat = molecule_data.MolData['BONDS']
    bdat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl1']]
    bdat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl2']]
    bnd_df = pd.DataFrame(bdat)
    bnd_df['KIJ'] = bnd_df['KIJ'] 
    bnd_df['RIJ'] = bnd_df['RIJ'] 
    bnd_df['UNQ'] = [pairing_func(i + 1, j + 1)[0]
                     for i, j in zip(bnd_df.cl1, bnd_df.cl2)]
    bnd_df['UNQ'] = [pairing_func(i, j)[0]
                     for i, j in zip(bnd_df.cl1, bnd_df.cl2)]
    bnd_df['UF'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl2
    bnd_df['UR'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl1
    connects = []
    for (ai, aj) in zip(bnd_df.cl1, bnd_df.cl2):
        connects.append("{:>5}{:>5}    1".format(aj + 1, ai + 1))
    full_bnd = bnd_df.copy()
    bnd_df = bnd_df.drop_duplicates(['TIJ'])
    return full_bnd, connects


def boss2gmxAngle(anglefile, num2opls, st_no, itpf):
    adat = anglefile
    adat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl1']]
    adat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl2']]
    adat['cl3'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl3']]
    ang_df = pd.DataFrame(adat)
    ang_df = ang_df[ang_df.K > 0]
    full_ang = ang_df.copy()
    ang_df['TY'] = np.array([num2opls[i] + '-' + num2opls[j] + '-' + num2opls[k]
                             for i, j, k in zip(ang_df.cl1, ang_df.cl2, ang_df.cl3)])
    ang_df = ang_df.drop_duplicates(['TY'])
    return full_ang


def bossData(molecule_data):
    ats_file = molecule_data.MolData['ATOMS']
    types = []
    for i in enumerate(ats_file):
        types.append([i[1].split()[1], 'opls_' + i[1].split()[2]])
    st_no = 3
    Qs = molecule_data.MolData['Q_LJ']
    assert len(Qs) == len(types), 'Please check the at_info and Q_LJ_dat files'
    num2opls = {}
    for i in range(0, len(types)):
        num2opls[i] = Qs[i][0]
    num2typ2symb = {i: types[i] for i in range(len(Qs))}
    for i in range(len(Qs)):
        num2typ2symb[i].append(bossPdbAtom2Element(
            num2typ2symb[i][0]) + num2typ2symb[i][1][-3:])
        num2typ2symb[i].append(bossPdbAtom2Element(num2typ2symb[i][0]))
        num2typ2symb[i].append(bossElement2Mass(num2typ2symb[i][3]))
        num2typ2symb[i].append(Qs[i][0])
    return (types, Qs, num2opls, st_no, num2typ2symb)


def boss2desmond(resid, molecule_data, pdb_file):
    types, Qs, num2opls, st_no, num2typ2symb = bossData(
        molecule_data)
    itpf = open(resid + '.cms', 'w+')
    itpf.write("""{
  s_m_m2io_version
  :::
  2.0.0
}
f_m_ct {
  s_m_title
  r_chorus_box_ax
  r_chorus_box_ay
  r_chorus_box_az
  r_chorus_box_bx
  r_chorus_box_by
  r_chorus_box_bz
  r_chorus_box_cx
  r_chorus_box_cy
  r_chorus_box_cz
  s_ffio_ct_type
  :::
  "Desmond file created by LigParGen (written by Leela S Dodda)"
                  70.0
                   0.0
                   0.0
                   0.0
                  70.0
                   0.0
                   0.0
                   0.0
                  70.0
  full_system
  m_atom[%d] {
    # First column is atom index #
    i_m_mmod_type
    r_m_x_coord
    r_m_y_coord
    r_m_z_coord
    i_m_residue_number
    s_m_pdb_residue_name
    i_m_atomic_number
    s_m_atom_name
    r_ffio_x_vel
    r_ffio_y_vel
    r_ffio_z_vel
    :::
"""%(len(Qs)))
    #print(molecule_data.MolData['XYZ'])
    xyz = molecule_data.MolData['XYZ']
    for i,r in xyz.iterrows(): 
        itpf.write('    %d        %d %10.8f %10.8f %10.8f      1 \"MOL\"     %d  \"%s\" 0.00000000 0.00000000 0.00000000\n'%(i+1,1,r.X,r.Y,r.Z,r.at_num,r.at_symb))
    bnd_df, connects = boss2gmxBond(molecule_data, st_no, itpf)
    itpf.write('''    :::
  }
  m_bond[%d] {
    i_m_from
    i_m_to
    i_m_order
    i_m_from_rep
    i_m_to_rep
    :::
'''%(len(bnd_df.index)))
    for rb in bnd_df.iterrows():
        i,r = rb
        itpf.write('    %d %d %d %d %d %d\n'%(i+1,r.cl2+1,r.cl1+1,1,1,1))
    itpf.write('''    :::
  }
}
f_m_ct {
  s_m_title
  r_chorus_box_ax
  r_chorus_box_ay
  r_chorus_box_az
  r_chorus_box_bx
  r_chorus_box_by
  r_chorus_box_bz
  r_chorus_box_cx
  r_chorus_box_cy
  r_chorus_box_cz
  s_ffio_ct_type
  :::
  "MOL"
                  70.0
                   0.0
                   0.0
                   0.0
                  70.0
                   0.0
                   0.0
                   0.0
                  70.0
  solute
  m_atom[%d] {
    # First column is atom index #
    i_m_mmod_type
    r_m_x_coord
    r_m_y_coord
    r_m_z_coord
    i_m_residue_number
    s_m_pdb_residue_name
    i_m_atomic_number
    s_m_atom_name
    r_ffio_x_vel
    r_ffio_y_vel
    r_ffio_z_vel
    :::
'''%(len(Qs)))
    xyz = molecule_data.MolData['XYZ']
    for i,r in xyz.iterrows():
        itpf.write('    %d        %d %10.8f %10.8f %10.8f      1 \"MOL\"     %d  \"%s\" 0.00000000 0.00000000 0.00000000\n'%(i+1,1,r.X,r.Y,r.Z,r.at_num,r.at_symb))
#    bnd_df, connects = boss2gmxBond(molecule_data, st_no, itpf)
    itpf.write('''    :::
  }
  m_bond[%d] {
    i_m_from
    i_m_to
    i_m_order
    i_m_from_rep
    i_m_to_rep
    :::
'''%(len(bnd_df.index)))
    for rb in bnd_df.iterrows():
        i,r = rb
        itpf.write('    %d %d %d %d %d %d\n'%(i+1,r.cl2+1,r.cl1+1,1,1,1))
#    for i,r in bnd_df.iterrows():itpf.write('    %d %d %d %d %d %d\n'%(i+1,r.cl2+1,r.cl1+1,1,1,1))
    itpf.write('''    :::
  }
  ffio_ff {
    s_ffio_name
    s_ffio_comb_rule
    i_ffio_version
    :::
    "MOL"
    GEOMETRIC
    1.0.0
    ffio_vdwtypes[%d] {
      s_ffio_name
      s_ffio_funct
      r_ffio_c1
      r_ffio_c2
      :::
'''%(len(Qs)))
    for i in range(len(Qs)): itpf.write('      %d %s LJ12_6_sig_epsilon %10.8f %10.8f\n'%(i+1,num2typ2symb[i][1],float(Qs[i][2]), float(Qs[i][3]))) 
    itpf.write('''      :::
    }
    ffio_sites[%d] {
      s_ffio_type
      r_ffio_charge
      r_ffio_mass
      s_ffio_vdwtype
      i_ffio_resnr
      s_ffio_residue
      :::
'''%(len(Qs)))
    for i in range(len(Qs)): itpf.write('     %d  atom %10.8f %10.8f %s 1  MOL\n'%(i+1,float(Qs[i][1]),num2typ2symb[i][4],num2typ2symb[i][1]))
    itpf.write('''      :::
    }''')
    itpf.write('''    ffio_bonds[%d] {
      i_ffio_ai
      i_ffio_aj
      s_ffio_funct
      r_ffio_c1
      r_ffio_c2
      :::
'''%(len(bnd_df.index)))
    EXCLUSIONS_12_13 = {'BI':[],'BJ':[],'UID':[]} 
    for rb in bnd_df.iterrows():
        i,r = rb
        itpf.write('      %d %d %d HARM      %10.8f    %10.8f\n'%(i+1,r.cl2+1,r.cl1+1,r.RIJ,r.KIJ))
        EXCLUSIONS_12_13['BI'].append(int(r.cl1+1));
        EXCLUSIONS_12_13['BJ'].append(int(r.cl2+1)); 
        EXCLUSIONS_12_13['UID'].append(pairing_func(r.cl2+1,r.cl1+1)[0])
    full_ang = boss2gmxAngle(molecule_data.MolData['ANGLES'], num2opls, st_no, itpf)
    itpf.write('''      :::
    }
    ffio_angles[%d] {
      i_ffio_ai
      i_ffio_aj
      i_ffio_ak
      s_ffio_funct
      r_ffio_c1
      r_ffio_c2
      ::: 
'''%(len(full_ang.index)))
    for rb in full_ang.iterrows():
        i,r = rb
        itpf.write('      %d %d %d %d HARM      %10.8f    %10.8f\n'%(i+1,r.cl1+1,r.cl2+1,r.cl3+1,r.R,r.K))
        EXCLUSIONS_12_13['BI'].append(int(r.cl1+1));
        EXCLUSIONS_12_13['BJ'].append(int(r.cl3+1)); 
        EXCLUSIONS_12_13['UID'].append(pairing_func(r.cl3+1,r.cl1+1)[0])
    full_tor, tor_df = boss2opmTorsion(bnd_df, num2opls, st_no, molecule_data, itpf)
    itpf.write('''      :::
    }
    ffio_dihedrals[%d] {
      i_ffio_ai
      i_ffio_aj
      i_ffio_ak
      i_ffio_al
      s_ffio_funct
      r_ffio_c0
      r_ffio_c1
      r_ffio_c2
      r_ffio_c3
      r_ffio_c4
      :::
'''%(len(full_tor.index)))
    full_tor.index = range(0,len(full_tor.index))
    EXCEPTION_14 = {'BI':[],'BJ':[],'UID':[]}
    for rb in full_tor.iterrows():
        i,r = rb
        if r.TY == 'Proper': 
            itpf.write('      %d %d %d %d %d Opls_proper  0   %10.8f    %10.8f       %10.8f    %10.8f\n'%(i+1,r.I+1,r.J+1,r.K+1,r.L+1,r.V1,r.V2,r.V3,r.V4))
            EXCLUSIONS_12_13['BI'].append(int(r.I+1));
            EXCLUSIONS_12_13['BJ'].append(int(r.L+1)); 
            EXCLUSIONS_12_13['UID'].append(pairing_func(r.I+1,r.L+1)[0])
            EXCEPTION_14['BI'].append(int(r.I+1));
            EXCEPTION_14['BJ'].append(int(r.L+1)); 
            EXCEPTION_14['UID'].append(pairing_func(r.I+1,r.L+1)[0])
        elif r.TY == 'Improper': itpf.write('      %d %d %d %d %d Opls_improper  0    %10.8f    %10.8f       %10.8f    %10.8f\n'%(i+1,r.I+1,r.J+1,r.K+1,r.L+1,r.V1,r.V2,r.V3,r.V4))
    excl = (pd.DataFrame(EXCLUSIONS_12_13))
    excl = excl.drop_duplicates(['UID'])
    excp = (pd.DataFrame(EXCEPTION_14))
    excp = excp.drop_duplicates(['UID'])
    itpf.write('''      :::
    }
    ffio_torsion_torsion[0] {
      :::
      :::
    }
''')
    itpf.write('''    ffio_exclusions[%d] {
      i_ffio_ai
      i_ffio_aj
      :::
'''%(len(excl.index)))
    excl.index = range(len(excl.index))
    for rb in excl.iterrows():
        i,r =rb
        itpf.write('      %d %d %d\n'%(i+1,r.BI,r.BJ))
    itpf.write('''      :::
    }
    ffio_pairs[%d] {
      i_ffio_ai
      i_ffio_aj
      s_ffio_funct
      r_ffio_c1
      r_ffio_c2
      :::
'''%(len(excp.index)*2))
#    prop_tor = full_tor[full_tor.TY=='Proper']
    excp.index = range(len(excp.index))
    for rb in excp.iterrows():
        i,r = rb
        itpf.write('      %d %d %d Coulomb 0.5 <>\n'%(2*i+1,r.BI,r.BJ))
        itpf.write('      %d %d %d LJ 0.5 <>\n'%(2*(i+1),r.BI,r.BJ))
    itpf.write('''      :::
    }
    ffio_constraints[0] {
      :::
      :::
    }
  }
}
''')
    itpf.close()
    return None


def mainBOSS2DESMOND(resid, clu):
    mol = pickle.load(open(resid + ".p", "rb"))
    if clu:
        pdb_file = 'clu.pdb'
    else:
        pdb_file = 'plt.pdb'
    boss2desmond(resid, mol, pdb_file)
    return None



"""
SCRIPT TO CONVERT WRITE CNS/X-PLOR PARAM AND TOP FILES 
FROM BOSS ZMATRIX
Created on Mon Feb 15 15:40:05 2016
@author: Leela S. Dodda leela.dodda@yale.edu
@author: William L. Jorgensen Lab 

REQUIREMENTS:
BOSS (need to set BOSSdir in bashrc and cshrc)
Preferably Anaconda python with following modules
pandas 
argparse
numpy
"""

from LigParGen.BOSSReader import bossPdbAtom2Element,bossElement2Mass,ucomb,tor_cent
import pickle
import pandas as pd
import numpy as np


def retDihedImp(df):
    odihed = []
    if np.sum([df['V' + str(pot)] for pot in range(1, 5)]) != 0.0:
        for pot in range(1, 5):
            if (df['V' + str(pot)] != 0.0):
                odihed.append('IMPRoper %s %4.5f %d %4.5f \n' % (df['NAME'].replace(
                    "-", "   "), df['V' + str(pot)], pot, 180.00 * abs(pot % 2 - 1)))
    else:
        pot = 2
        odihed.append('IMPRoper %s %4.5f %d %4.5f \n' % (df['NAME'].replace(
            "-", "   "), df['V' + str(pot)], pot, 180.00 * abs(pot % 2 - 1)))
    return (odihed)


def retDihed(df):
    odihed = []
    odihed.append('DIHEdral %s  MULT 4   %4.3f     %d   %4.2f \n' % (df['NAME'].replace("-", "   "), df['V' + str(1)], 1, 180.00 * abs(1 % 2 - 1)))
    for pot in range(2, 5):
        odihed.append('                                             %4.3f     %d   %4.2f \n' % (df['V' + str(pot)], pot, 180.00 * abs(pot % 2 - 1)))
    return (odihed)


def Boss2CharmmRTF(num2typ2symb, Qs, resid, bnd_df, angs,props,imps):
    charges = [float(Qs[i][1]) for i in range(len(Qs))]
    rtf = open(resid + '.top', 'w+')
    rtf.write('Remarks LigParGen generated XPLOR-TOP file for Bonvin group (by Leela Dodda)\n')
    rtf.write('\nset echo=false end\n')
    rtf.write('\nautogenerate angles=True dihedrals=True end\n')
    rtf.write('{ atomType  mass }\n')
    Mass = ['MASS %s %3.4f \n' % (num2typ2symb[i][2], bossElement2Mass(bossPdbAtom2Element(num2typ2symb[i][0]))) for i in range(len(Qs))]
    for i in range(len(Mass)):
        rtf.write('%s' % Mass[i])
    rtf.write('\nRESIdue %5s\n' % (resid))
    rtf.write('\nGROUP\n')
    rtf.write('\n{ atomName  atomType  Charge } \n')
    for i in range(len(Qs)):
        rtf.write('ATOM %6s TYPE= %6s CHARGE= %8s END\n' % (num2typ2symb[i][0], num2typ2symb[i][2], Qs[i][1]))
    rtf.write('\n{ Bonds: atomName1  atomName2 } \n')
    for (x, y) in zip(bnd_df.cl1, bnd_df.cl2):
        rtf.write('BOND %s %s \n' % (num2typ2symb[x][0], num2typ2symb[y][0]))
    bndlist = list(bnd_df.UR) + list(bnd_df.UR)
    imp_list = []
    for i,dat in imps.iterrows():
        ndata =tor_cent([dat.I, dat.J, dat.K, dat.L], bndlist) 
        sdata = [num2typ2symb[j][0] for j in ndata]
        imp_list.append('-'.join(sdata))
    rtf.write('\n{ Improper Dihedrals: aName1 aName2 aName3 aName4 }\n')
    for i in imp_list:
        rtf.write('IMPRoper %s \n' % (i.replace("-", " ")))
    rtf.write('\nEND {RESIdue UNK}\n')
    rtf.write('\nset echo=true end\n')
    rtf.close()
    return None


def Boss2CharmmPRM(resid, num2typ2symb, Qs, bnd_df, ang_df, tor_df):
    #### COLLECTING NONBONDING PART #######
    prm = open(resid + '.param', 'w+')
    prm.write('Remarks LigParGen generated XPLOR-TOP file for Bonvin group (by Leela Dodda)\n')
    prm.write('\nset echo=false end \n')
    prm.write('\n{ Bonds: atomType1 atomType2 kb r0 } \n')
    for i in bnd_df.index:
        prm.write('BOND %6s %6s %8.1f %8.4f \n' % (num2typ2symb[bnd_df.cl1[i]][
            2], num2typ2symb[bnd_df.cl2[i]][2], bnd_df.KIJ[i], bnd_df.RIJ[i]))
    prm.write('\n{ Angles: aType1 aType2 aType3 kt t0 }\n')
    for i in ang_df.index:
        prm.write('ANGLe %5s %5s %5s %8.1f %8.2f \n' % (num2typ2symb[ang_df.cl1[i]][2], num2typ2symb[
            ang_df.cl2[i]][2], num2typ2symb[ang_df.cl3[i]][2], ang_df.K[i], ang_df.R[i]))
    prm.write('\n{ Proper Dihedrals: aType1 aType2 aType3 aType4 kt period phase } \n')
    if len(tor_df.index) > 0:
        tor_df = tor_df.drop_duplicates(['NAME', 'TY'])
    pro_df = tor_df[tor_df.TY == 'Proper']
    for i in list(pro_df.index):
        ndf = pro_df.ix[i]
        pro_out = retDihed(ndf.to_dict())
        for i in range(4):
            prm.write('%s' % pro_out[i])
    prm.write(
        'DIHEdral    X      X      X      X  MULT 1   0.000     1     0.00 ! WILD CARD FOR MISSING TORSION PARAMETERS\n')
    prm.write('\n{ Improper Dihedrals: aType1 aType2 aType3 aType4 kt period phase }\n')
    imp_df = tor_df[tor_df.TY == 'Improper']
    for i in list(imp_df.index):
        ndf = tor_df.ix[i]
        imp_out = retDihedImp(ndf.to_dict())
        for i in range(len(imp_out)):
            prm.write('%s' % imp_out[i])
    prm.write(
        'IMPRoper    X      X      X      X  0.000     1     0.00  ! WILD CARD FOR MISSING IMPROPER PARAMETERS \n')
    prm.write(
        '\n{ Nonbonded: Type Emin sigma; (1-4): Emin/2 sigma }\n')
    Qlines = ['NONBonded %5s %11.6f %11.6f %11.6f %11.6f \n' %
              (num2typ2symb[i][2], float(Qs[i][3]), float(Qs[i][2]), float(Qs[i][3]) * 0.50,float(Qs[i][2])) for i in range(len(Qs))]
    for i in range(len(Qlines)):
        prm.write('%s' % Qlines[i])
    prm.write('\nset echo=true end \n')
    prm.close()
    return None


def Boss2CharmmTorsion(bnd_df, num2opls, st_no, molecule_data, num2typ2symb):
    dhd = []
    for line in molecule_data.MolData['TORSIONS']:
        dt = [float(l) for l in line]
        dhd.append(dt)
    dhd = np.array(dhd)
    dhd = dhd  # kcal to kj conversion
    dhd = dhd / 2.0  # Komm = Vopls/2
    dhd_df = pd.DataFrame(dhd, columns=['V1', 'V2', 'V3', 'V4'])
    ats = []
    for line in molecule_data.MolData['ATOMS'][3:]:
        dt = [line.split()[0], line.split()[4],
              line.split()[6], line.split()[8]]
        dt = [int(d) for d in dt]
        ats.append(dt)
    for line in molecule_data.MolData['ADD_DIHED']:
        dt = [int(l) for l in line]
        ats.append(dt)
    assert len(ats) == len(
        dhd), 'Number of Dihedral angles in Zmatrix and Out file dont match'
    ats = np.array(ats) - st_no
    for i in range(len(ats)):
        for j in range(len(ats[0])):
            if ats[i][j] < 0:
                ats[i][j] = 0
    at_df = pd.DataFrame(ats, columns=['I', 'J', 'K', 'L'])
    final_df = pd.concat([dhd_df, at_df], axis=1, join_axes=[at_df.index])
    bndlist = list(bnd_df.UR) + (list(bnd_df.UR))
    final_df['TY'] = ['Proper' if ucomb(list([final_df.I[n], final_df.J[n], final_df.K[
        n], final_df.L[n]]), bndlist) == 3 else 'Improper' for n in range(len(final_df.I))]
    final_df['TI'] = [num2typ2symb[j][2] for j in final_df.I]
    final_df['TJ'] = [num2typ2symb[j][2] for j in final_df.J]
    final_df['TK'] = [num2typ2symb[j][2] for j in final_df.K]
    final_df['TL'] = [num2typ2symb[j][2] for j in final_df.L]
    final_df['SYMB'] = ['   '.join([num2typ2symb[final_df.I[i]][0], num2typ2symb[final_df.J[i]][
        0], num2typ2symb[final_df.K[i]][0], num2typ2symb[final_df.L[i]][0]]) for i in final_df.index]
    if len(final_df.index) > 0:
        final_df['NAME'] = final_df.TI + '-' + final_df.TJ + \
            '-' + final_df.TK + '-' + final_df.TL
    return final_df


def boss2CharmmBond(molecule_data, st_no):
    bdat = molecule_data.MolData['BONDS']
    bdat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl1']]
    bdat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl2']]
    bnd_df = pd.DataFrame(bdat)
    bnd_df['UF'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl2
    bnd_df['UR'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl1
#    bnd_df.to_csv('bos_bonds.csv', index=False)
    hb_df = bnd_df.drop(['cl1', 'cl2', 'UF', 'UR'], 1)
    hb_df = hb_df.drop_duplicates()
    return bnd_df


def boss2CharmmAngle(anglefile, num2opls, st_no,num2typ2symb):
    adat = anglefile
    adat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl1']]
    adat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl2']]
    adat['cl3'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl3']]
    ang_df = pd.DataFrame(adat)
    ang_df = ang_df[ang_df.K > 0]
#    ang_df.to_csv('bos_angles.csv', index=False)
    ang_df['TY'] = np.array([num2opls[i] + '-' + num2opls[j] + '-' + num2opls[k]
                             for i, j, k in zip(ang_df.cl1, ang_df.cl2, ang_df.cl3)])
    ang_df['TI']=[num2typ2symb[ang_df.cl1[i]][2] for i in ang_df.index]
    ang_df['TJ']=[num2typ2symb[ang_df.cl2[i]][2] for i in ang_df.index]
    ang_df['TK']=[num2typ2symb[ang_df.cl3[i]][2] for i in ang_df.index]
    ang_df['TY'] = np.array([i + '  ' + j + '  ' + k
                             for i, j, k in zip(ang_df.TI, ang_df.TJ, ang_df.TK)])
    return ang_df


def bossData(molecule_data):
    ats_file = molecule_data.MolData['ATOMS']
    types = []
    for i in enumerate(ats_file):
        types.append([i[1].split()[1], 'opls_' + i[1].split()[2]])
    st_no = 3
    Qs = molecule_data.MolData['Q_LJ']
    assert len(Qs) == len(types), 'Please check the at_info and Q_LJ_dat files'
    num2opls = {}
    for i in range(0, len(types)):
        num2opls[i] = Qs[i][0]
    num2typ2symb = {i: types[i] for i in range(len(Qs))}
    for i in range(len(Qs)):
        trm = bossPdbAtom2Element(num2typ2symb[i][0]) + num2typ2symb[i][1][-3:]
        num2typ2symb[i].append(trm[0:4])
        num2typ2symb[i].append(bossPdbAtom2Element(num2typ2symb[i][0]))
        num2typ2symb[i].append(bossElement2Mass(num2typ2symb[i][3]))
        num2typ2symb[i].append(Qs[i][0])
    return (types, Qs, num2opls, st_no, num2typ2symb)


def Boss2Charmm(resid, molecule_data):
    types, Qs, num2opls, st_no, num2typ2symb = bossData(molecule_data)
    bnd_df = boss2CharmmBond(molecule_data, st_no)
    ang_df = boss2CharmmAngle(molecule_data.MolData['ANGLES'], num2opls, st_no,num2typ2symb)
    tor_df = Boss2CharmmTorsion(bnd_df, num2opls, st_no,
                                molecule_data, num2typ2symb)
    Boss2CharmmRTF(num2typ2symb, Qs, resid, bnd_df, list(ang_df['TY']), list(
        tor_df[tor_df.TY == 'Proper']['SYMB']),tor_df[tor_df.TY == 'Improper'])
    Boss2CharmmPRM(resid, num2typ2symb, Qs, bnd_df, ang_df, tor_df)
    return None


def mainBOSS2XPLOR(resid, clu=False):
    mol = pickle.load(open(resid + ".p", "rb"))
    Boss2Charmm(resid, mol)
    return None



"""
SCRIPT TO  WRITE LAMMPS DATA FILES LMP & INP Files 
FROM BOSS ZMATRIX
Created on Mon Sep 30 03:31:05 2017
@author: Leela S. Dodda leela.dodda@yale.edu
@author: William L. Jorgensen Lab 

REQUIREMENTS:
BOSS (need to set BOSSdir in bashrc and cshrc)
Preferably Anaconda python with following modules
pandas 
argparse
numpy
"""

from LigParGen.BOSSReader import bossPdbAtom2Element,bossElement2Mass,ucomb,tor_cent
import pickle
import pandas as pd
import numpy as np


def Boss2LammpsLMP(resid, num2typ2symb, Qs, bnd_df, ang_df, tor_df,molecule_data):
    xyz_df = molecule_data.MolData['XYZ']
    max_mol_size = 50 
    prm = open(resid + '.lmp', 'w+')
    prm.write('LAMMPS data file Created by LigParGen - (Written by Leela S. Dodda)\n\n')
    prm.write('%8d atoms\n'%len(Qs))
    prm.write('%8d bonds\n'%len(bnd_df.KIJ))
    prm.write('%8d angles\n'%len(ang_df.K))
    prm.write('%8d dihedrals\n'%len(tor_df[tor_df.TY == 'Proper'].index))
    prm.write('%8d impropers\n \n'%len(tor_df[tor_df.TY == 'Improper'].index))
    prm.write('%8d atom types\n'%len(Qs))
    prm.write('%8d bond types\n'%len(bnd_df.KIJ))
    prm.write('%8d angle types\n'%len(ang_df.K))
    prm.write('%8d dihedral types\n'%len(tor_df[tor_df.TY == 'Proper'].index))
    prm.write('%8d improper types\n \n'%len(tor_df[tor_df.TY == 'Improper'].index))
    prm.write('%12.6f %12.6f xlo xhi\n'%(xyz_df.X.min(),xyz_df.X.min()+max_mol_size))
    prm.write('%12.6f %12.6f ylo yhi\n'%(xyz_df.Y.min(),xyz_df.Y.min()+max_mol_size))
    prm.write('%12.6f %12.6f zlo zhi\n'%(xyz_df.Z.min(),xyz_df.Z.min()+max_mol_size))
    #### Printing Parameters for ALL BONDS/ANGLES/DIHEDRALS/IMPROPERS/Q/LJ #######
    prm.write('\nMasses\n\n')
    for i in range(len(Qs)): prm.write('%8d %10.3f  \n'%(i+1,float(num2typ2symb[i][4])))
    prm.write('\nPair Coeffs \n\n')
    for i in range(len(Qs)): prm.write('%8d%11.3f%11.7f \n'%(i+1,float(Qs[i][3]),float(Qs[i][2])))
    prm.write('\nBond Coeffs \n\n')
    for i in bnd_df.index:
        prm.write('%8d%11.4f%11.4f \n' % (i+1, bnd_df.KIJ[i], bnd_df.RIJ[i]))
    prm.write('\nAngle Coeffs \n\n')
    for i in ang_df.index:
        prm.write('%8d%11.3f%11.3f\n' % (i+1, ang_df.K[i],ang_df.R[i]))
    dihedral_df = tor_df[tor_df.TY == 'Proper'] 
    dihedral_df.index = range(len(dihedral_df.V1))
    prm.write('\nDihedral Coeffs  \n\n')
    for i,row in dihedral_df.iterrows(): 
        prm.write('%8d%11.3f%11.3f%11.3f%11.3f \n'%(i+1,row.V1,row.V2,row.V3,row.V4))
    bndlist = list(bnd_df.UR) + (list(bnd_df.UR))
    improper_df = tor_df[tor_df.TY == 'Improper']
    improper_df.index = range(len(improper_df.V2))
    if len(improper_df.index)>0:
        prm.write('\nImproper Coeffs \n\n')
        for i,row in  improper_df.iterrows(): prm.write('%8d%11.3f%8d%8d \n'%(i+1,row.V2*0.5,-1,2))
    #### Printing EXPLICITLY ALL BONDS/ANGLES/DIHEDRALS/IMPROPERS/Q/LJ #######
    prm.write('\nAtoms \n\n')
    for i in range(len(xyz_df.index)):prm.write('%6d %6d %6d %10.8f %8.3f %8.5f %8.5f\n'%(i+1,1,i+1,float(Qs[i][1]),xyz_df.X[i],xyz_df.Y[i],xyz_df.Z[i]))
    prm.write('\nBonds \n\n')
    for i in bnd_df.index:
        prm.write('%6d %6d %6d %6d\n' % (i+1, i+1,bnd_df.cl1[i]+1, bnd_df.cl2[i]+1))
    prm.write('\nAngles \n\n')
    for i in ang_df.index:
        prm.write('%6d %6d %6d %6d %6d\n' % (i+1, i+1, ang_df.cl1[i]+1, ang_df.cl2[i]+1,ang_df.cl3[i]+1))
    prm.write('\nDihedrals\n\n')
    for i,row in dihedral_df.iterrows(): 
        prm.write('%6d %6d %6d %6d %6d %6d \n'%(i+1,i+1,row.I+1,row.J+1,row.K+1,row.L+1))
    if len(improper_df.index)>0:
        prm.write('\nImpropers\n\n')
        for row in improper_df.iterrows():
            index, dat = row
            ndata = tor_cent([dat.I, dat.J, dat.K, dat.L], bndlist)
            prm.write('%6d %6d %6d %6d %6d %6d \n'%(index+1,index+1,ndata[0]+1,ndata[1]+1,ndata[2]+1,ndata[3]+1))
    return None


def Boss2CharmmTorsion(bnd_df, num2opls, st_no, molecule_data, num2typ2symb):
    dhd = []
    for line in molecule_data.MolData['TORSIONS']:
        dt = [float(l) for l in line]
        dhd.append(dt)
    dhd = np.array(dhd)
    dhd = dhd  # kcal to kj conversion
    dhd = dhd # Klammps = Vopls
    dhd_df = pd.DataFrame(dhd, columns=['V1', 'V2', 'V3', 'V4'])
    ats = []
    for line in molecule_data.MolData['ATOMS'][3:]:
        dt = [line.split()[0], line.split()[4],
              line.split()[6], line.split()[8]]
        dt = [int(d) for d in dt]
        ats.append(dt)
    for line in molecule_data.MolData['ADD_DIHED']:
        dt = [int(l) for l in line]
        ats.append(dt)
    assert len(ats) == len(
        dhd), 'Number of Dihedral angles in Zmatrix and Out file dont match'
    ats = np.array(ats) - st_no
    for i in range(len(ats)):
        for j in range(len(ats[0])):
            if ats[i][j] < 0:
                ats[i][j] = 0
    at_df = pd.DataFrame(ats, columns=['I', 'J', 'K', 'L'])
    final_df = pd.concat([dhd_df, at_df], axis=1, join_axes=[at_df.index])
    bndlist = list(bnd_df.UR) + (list(bnd_df.UR))
    final_df['TY'] = ['Proper' if ucomb(list([final_df.I[n], final_df.J[n], final_df.K[
        n], final_df.L[n]]), bndlist) == 3 else 'Improper' for n in range(len(final_df.I))]
    final_df['TI'] = [num2typ2symb[j][2] for j in final_df.I]
    final_df['TJ'] = [num2typ2symb[j][2] for j in final_df.J]
    final_df['TK'] = [num2typ2symb[j][2] for j in final_df.K]
    final_df['TL'] = [num2typ2symb[j][2] for j in final_df.L]
    final_df['SYMB'] = ['-'.join([num2typ2symb[final_df.I[i]][0], num2typ2symb[final_df.J[i]][
        0], num2typ2symb[final_df.K[i]][0], num2typ2symb[final_df.L[i]][0]]) for i in final_df.index]
    if len(final_df.index) > 0:
        final_df['NAME'] = final_df.TI + '-' + final_df.TJ + \
            '-' + final_df.TK + '-' + final_df.TL
    return final_df


def boss2CharmmBond(molecule_data, st_no):
    bdat = molecule_data.MolData['BONDS']
    bdat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl1']]
    bdat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in bdat['cl2']]
    bnd_df = pd.DataFrame(bdat)
    bnd_df['UF'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl2
    bnd_df['UR'] = ((bnd_df.cl1 + bnd_df.cl2) *
                    (bnd_df.cl1 + bnd_df.cl2 + 1) * 0.5) + bnd_df.cl1
    hb_df = bnd_df.drop(['cl1', 'cl2', 'UF', 'UR'], 1)
    hb_df = hb_df.drop_duplicates()
    return bnd_df


def boss2CharmmAngle(anglefile, num2opls, st_no):
    adat = anglefile
    adat['cl1'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl1']]
    adat['cl2'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl2']]
    adat['cl3'] = [x - st_no if not x - st_no < 0 else 0 for x in adat['cl3']]
    ang_df = pd.DataFrame(adat)
    ang_df = ang_df[ang_df.K > 0]
    ang_df['TY'] = np.array([num2opls[i] + '-' + num2opls[j] + '-' + num2opls[k]
                             for i, j, k in zip(ang_df.cl1, ang_df.cl2, ang_df.cl3)])
    return ang_df


def bossData(molecule_data):
    ats_file = molecule_data.MolData['ATOMS']
    types = []
    for i in enumerate(ats_file):
        types.append([i[1].split()[1], 'opls_' + i[1].split()[2]])
    st_no = 3
    Qs = molecule_data.MolData['Q_LJ']
    assert len(Qs) == len(types), 'Please check the at_info and Q_LJ_dat files'
    num2opls = {}
    for i in range(0, len(types)):
        num2opls[i] = Qs[i][0]
    num2typ2symb = {i: types[i] for i in range(len(Qs))}
    for i in range(len(Qs)):
        num2typ2symb[i].append(bossPdbAtom2Element(
            num2typ2symb[i][0]) + num2typ2symb[i][1][-3:])
        num2typ2symb[i].append(bossPdbAtom2Element(num2typ2symb[i][0]))
        num2typ2symb[i].append(bossElement2Mass(num2typ2symb[i][3]))
        num2typ2symb[i].append(Qs[i][0])
    return (types, Qs, num2opls, st_no, num2typ2symb)


def Boss2Lammps(resid, molecule_data):
    types, Qs, num2opls, st_no, num2typ2symb = bossData(molecule_data)
    bnd_df = boss2CharmmBond(molecule_data, st_no)
    ang_df = boss2CharmmAngle(molecule_data.MolData['ANGLES'], num2opls, st_no)
    tor_df = Boss2CharmmTorsion(bnd_df, num2opls, st_no,
                                molecule_data, num2typ2symb)
    Boss2LammpsLMP(resid, num2typ2symb, Qs, bnd_df, ang_df, tor_df,molecule_data)
    return None


def mainBOSS2LAMMPS(resid, clu=False):
    mol = pickle.load(open(resid + ".p", "rb"))
    Boss2Lammps(resid, mol)
    return None
